// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;

/** a parser which translates an Input&lt;Token&gt; into a Forest&lt;NodeType&gt; */
public abstract class Parser<Token, NodeType> {

    protected final Table<Token> pt;

    /** create a parser to parse the grammar with start symbol <tt>u</tt> */
    public Parser(Union u, Topology<Token> top)  { this.pt = new Table<Token>(u, top); }

    /** implement this method to create the output forest corresponding to a lone shifted input token */
    public abstract Forest<NodeType> shiftToken(Token t, Input.Location newloc);

    public String toString() { return pt.toString(); }

    private boolean verbose = false;;
    private static final char[] spin = new char[] { '-', '\\', '|', '/' };
    private int spinpos = 0;
    private long last = 0;
    void spin() {
        if (verbose) {
            long now = System.currentTimeMillis();
            if (now-last < 70) return;
            last = now;
            System.err.print("\r  " + spin[spinpos++ % (spin.length)]+"\r");
        }
    }

    /** parse <tt>input</tt>, and return the shared packed parse forest (or throw an exception) */
    public Forest<NodeType> parse(Input<Token> input) throws IOException, ParseFailed {
        verbose = System.getProperty("sbp.verbose", null) != null;
        spinpos = 0;
        try {
            GSS gss = new GSS(input, this);
            for(GSS.Phase current = gss.new Phase<Token>(pt.start); ;) {
                
                if (verbose) {
                    String s;
                    s = "  " + spin[spinpos++ % (spin.length)]+" parsing ";
                    s += input.getName();
                    s += " "+input.getLocation();
                    while(s.indexOf(':') != -1 && s.indexOf(':') < 8) s = " " + s;
                    String y = "@"+gss.viewPos+" ";
                    while(y.length() < 9) y = " " + y;
                    s += y;
                    //s += "   doom="+Node.doomedNodes;
                    //while(s.length() < 40) s = s + " ";
                    s += "   nodes="+gss.numOldNodes;
                    while(s.length() < 50) s = s + " ";
                    s += " shifted="+gss.numNewNodes;
                    while(s.length() < 60) s = s + " ";
                    s += " reductions="+gss.numReductions;
                    System.err.print("\r"+s+ANSI.clreol()+"\r");
                }
                
                // FIXME: make sure all the locations line up properly in here
                if (current.isDone()) return (Forest<NodeType>)current.finalResult;
                Forest forest = shiftToken((Token)current.token, input.getLocation());
                current = gss.new Phase<Token>(current, forest);
            }
        } finally {
            if (verbose)
                System.err.print("\r                                                                                \r");
        }
    }


    // Table //////////////////////////////////////////////////////////////////////////////

    /** an SLR(1) parse table which may contain conflicts */
    static class Table<Token> extends Cache {

        /** the start state */
        public  final State<Token>   start;

        /** the state from which no reductions can be done */
        private final State<Token>   dead_state;

        /** used to generate unique values for State.idx */
        private int master_state_idx = 0;
        HashSet<State<Token>>   all_states    = new HashSet<State<Token>>();
        HashMap<HashSet<Position>,State<Token>>   doomed_states    = new HashMap<HashSet<Position>,State<Token>>();
        HashMap<HashSet<Position>,State<Token>>   normal_states    = new HashMap<HashSet<Position>,State<Token>>();

        /** construct a parse table for the given grammar */
        public Table(Topology top) { this("s", top); }
        public Table(String startSymbol, Topology top) { this(new Union(startSymbol), top); }
        public Table(Union ux, Topology top) {
            super(ux, top);
            Union start0 = new Union("0");
            Sequence seq0 = new Sequence.Singleton(ux);
            start0.add(seq0);
            buildFollowSet(seq0, top, true);

            // construct the set of states
            HashSet<Position> hp = new HashSet<Position>();
            reachable(start0, hp);

            this.dead_state = new State<Token>(new HashSet<Position>(), true);
            this.start = new State<Token>(hp);

            // for each state, fill in the corresponding "row" of the parse table
            for(State<Token> state : all_states)
                for(Position p : state.hs) {

                    // the Grammar's designated "last position" is the only accepting state
                    if (start0.contains(p.owner()) && p.next()==null && !state.doomed)
                        state.accept = true;

                    if (isRightNullable(p)) {
                        Topology<Token> follow = (Topology<Token>)follow(p.owner());
                        for(Position p2 = p; p2 != null && p2.element() != null; p2 = p2.next()) {
                            if (!(p2.element() instanceof Union)) throw new Error("impossible");
                            Union u = (Union)p2.element();
                            Atom set = new edu.berkeley.sbp.chr.CharAtom(new edu.berkeley.sbp.chr.CharTopology((Topology<Character>)epsilonFollowSet(u)));
                            Element p2e = p2.element();
                            if (p2e instanceof Union)
                                for(Sequence p2es : ((Union)p2e))
                                    follow = follow.intersect(follow(p2es));
                            if (set != null) follow = follow.intersect(set.getTokenTopology());
                        }
                        state.reductions.put(follow, p);
                        if (followEof.contains(p.owner())) state.eofReductions.add(p);
                    }

                    // if the element following this position is an atom, copy the corresponding
                    // set of rows out of the "master" goto table and into this state's shift table
                    if (p.element() != null && p.element() instanceof Atom)
                        state.shifts.addAll(state.gotoSetTerminals.subset(((Atom)p.element()).getTokenTopology()));
                }

            if (top instanceof IntegerTopology)
                for(State<Token> state : all_states) {
                    state.oreductions = state.reductions.optimize(((IntegerTopology)top).functor());
                    state.oshifts = state.shifts.optimize(((IntegerTopology)top).functor());
                }

            // crude algorithm to assing an ordinal ordering to every position
            // al will be sorted in DECREASING order (al[0] >= al[1])
            ArrayList<Sequence.Position> al = new ArrayList<Sequence.Position>();
            for(State s : all_states) {
                for(Object po : s) {
                    Sequence.Position p = (Sequence.Position)po;
                    if (al.contains(p)) continue;
                    int i=0;
                    for(; i<al.size(); i++) {
                        if (comparePositions(p, al.get(i)) < 0)
                            break;
                    }
                    al.add(i, p);
                }
            }
            // FIXME: this actually pollutes the "pure" objects (the ones that should not be modified by the Parser)
            // sort in increasing order...
            OUTER: while(true) {
                for(int i=0; i<al.size(); i++)
                    for(int j=i+1; j<al.size(); j++)
                        if (comparePositions(al.get(i), al.get(j)) > 0) {
                            Sequence.Position p = al.remove(j);
                            al.add(i, p);
                            continue OUTER;
                        }
                break;
            }

            int j = 1;
            int pk = 0;
            for(int i=0; i<al.size(); i++) {
                boolean inc = false;
                for(int k=pk; k<i; k++) {
                    if (comparePositions(al.get(k), al.get(i)) > 0)
                        { inc = true; break; }
                }
                inc = true;
                if (inc) {
                    j++;
                    pk = i;
                }
                al.get(i).ord = j;
            }

            /*
            for(int i=0; i<al.size(); i++)
                if (isRightNullable(al.get(i)))
                    System.out.println(al.get(i).ord + "   " + al.get(i));
            */
            //mastercache = this;
        }

        /** a single state in the LR table and the transitions possible from it */
        class State<Token> implements IntegerMappable, Iterable<Position> {
        
            public  final     int               idx    = master_state_idx++;
            private final     HashSet<Position> hs;
            public HashSet<State<Token>> also = new HashSet<State<Token>>();

            public  transient HashMap<Sequence,State<Token>>         gotoSetNonTerminals = new HashMap<Sequence,State<Token>>();
            private transient TopologicalBag<Token,State<Token>>     gotoSetTerminals    = new TopologicalBag<Token,State<Token>>();

            private           TopologicalBag<Token,Position>      reductions          = new TopologicalBag<Token,Position>();
            private           HashSet<Position>                   eofReductions       = new HashSet<Position>();
            private           TopologicalBag<Token,State<Token>>  shifts              = new TopologicalBag<Token,State<Token>>();
            private           boolean                             accept              = false;

            private VisitableMap<Token,State<Token>> oshifts     = null;
            private VisitableMap<Token,Position>     oreductions = null;

            // Interface Methods //////////////////////////////////////////////////////////////////////////////

            boolean                    isAccepting()           { return accept; }
            public Iterator<Position>  iterator()              { return hs.iterator(); }
            boolean                    canShift(Token t)       { return oshifts!=null && oshifts.contains(t); }
            void                       invokeShifts(Token t, GSS.Phase phase, Result r) { oshifts.invoke(t, phase, r); }
            boolean                    canReduce(Token t)        {
                return oreductions != null && (t==null ? eofReductions.size()>0 : oreductions.contains(t)); }
            void          invokeEpsilonReductions(Token t, Node node) {
                if (t==null) for(Position r : eofReductions) node.invoke(r, null);
                else         oreductions.invoke(t, node, null);
            }
            void          invokeReductions(Token t, Node node, Result b) {
                if (t==null) for(Position r : eofReductions) node.invoke(r, b);
                else         oreductions.invoke(t, node, b);
            }

            // Constructor //////////////////////////////////////////////////////////////////////////////

            /**
             *  create a new state consisting of all the <tt>Position</tt>s in <tt>hs</tt>
             *  @param hs           the set of <tt>Position</tt>s comprising this <tt>State</tt>
             *  @param all the set of all elements (Atom instances need not be included)
             *  
             *   In principle these two steps could be merged, but they
             *   are written separately to highlight these two facts:
             * <ul>
             * <li> Non-atom elements either match all-or-nothing, and do not overlap
             *      with each other (at least not in the sense of which element corresponds
             *      to the last reduction performed).  Therefore, in order to make sure we
             *      wind up with the smallest number of states and shifts, we wait until
             *      we've figured out all the token-to-position multimappings before creating
             *      any new states
             *  
             * <li> In order to be able to run the state-construction algorithm in a single
             *      shot (rather than repeating until no new items appear in any state set),
             *      we need to use the "yields" semantics rather than the "produces" semantics
             *      for non-Atom Elements.
             *  </ul>
             */
            public State(HashSet<Position> hs) { this(hs, false); }
            public boolean doomed;
            public State(HashSet<Position> hs, boolean doomed) {
                this.hs = hs;
                this.doomed = doomed;

                // register ourselves in the all_states hash so that no
                // two states are ever created with an identical position set
                ((HashMap)(doomed ? doomed_states : normal_states)).put(hs, this);
                ((HashSet)all_states).add(this);
                
                for(Position p : hs) {
                    if (!p.isFirst()) continue;
                    for(Sequence s : p.owner().needs()) {
                        if (hs.contains(s.firstp())) continue;
                        HashSet<Position> h2 = new HashSet<Position>();
                        reachable(s, h2);
                        also.add(mkstate(h2, true));
                    }
                    for(Sequence s : p.owner().hates()) {
                        if (hs.contains(s.firstp())) continue;
                        HashSet<Position> h2 = new HashSet<Position>();
                        reachable(s, h2);
                        also.add(mkstate(h2, true));
                    }
                }

                // Step 1a: examine all Position's in this state and compute the mappings from
                //          sets of follow tokens (tokens which could follow this position) to sets
                //          of _new_ positions (positions after shifting).  These mappings are
                //          collectively known as the _closure_

                TopologicalBag<Token,Position> bag0 = new TopologicalBag<Token,Position>();
                for(Position position : hs) {
                    if (position.isLast() || !(position.element() instanceof Atom)) continue;
                    Atom a = (Atom)position.element();
                    HashSet<Position> hp = new HashSet<Position>();
                    reachable(position.next(), hp);
                    bag0.addAll(a.getTokenTopology(), hp);
                }

                // Step 1b: for each _minimal, contiguous_ set of characters having an identical next-position
                //          set, add that character set to the goto table (with the State corresponding to the
                //          computed next-position set).

                for(Topology<Token> r : bag0) {
                    HashSet<Position> h = new HashSet<Position>();
                    for(Position p : bag0.getAll(r)) h.add(p);
                    ((TopologicalBag)gotoSetTerminals).put(r, mkstate(h, doomed));
                }

                // Step 2: for every Sequence, compute the closure over every position in this set which
                //         is followed by a symbol which could yield the Sequence.
                //
                //         "yields" [in one or more step] is used instead of "produces" [in exactly one step]
                //         to avoid having to iteratively construct our set of States as shown in most
                //         expositions of the algorithm (ie "keep doing XYZ until things stop changing").

                HashMapBag<Sequence,Position> move = new HashMapBag<Sequence,Position>();
                for(Position p : hs)
                    if (!p.isLast() && p.element() instanceof Union)
                        for(Sequence s : ((Union)p.element())) {
                            HashSet<Position> hp = new HashSet<Position>();
                            reachable(p.next(), hp);
                            move.addAll(s, hp);
                        }
                OUTER: for(Sequence y : move) {
                    // if a reduction is "lame", it should wind up in the dead_state after reducing
                    HashSet<Position> h = move.getAll(y);
                    State<Token> s = mkstate(h, doomed);
                    for(Position p : hs)
                        if (p.element() != null && (p.element() instanceof Union))
                            for(Sequence seq : ((Union)p.element()))
                                if (seq.needs.contains(y) || seq.hates.contains(y)) {
                                    // FIXME: assumption that no sequence is ever both usefully (non-lamely) matched
                                    //        and also directly lamely matched
                                    ((HashMap)gotoSetNonTerminals).put(y, dead_state);
                                    continue OUTER;
                                }
                    gotoSetNonTerminals.put(y, s);
                }
            }

            private State<Token> mkstate(HashSet<Position> h, boolean b) {
                if (b) return doomed_states.get(h) == null ? (State)new State<Token>(h,b) : (State)doomed_states.get(h);
                else   return normal_states.get(h) == null ? (State)new State<Token>(h,b) : (State)normal_states.get(h);
            }

            public String toStringx() {
                StringBuffer st = new StringBuffer();
                for(Position p : this) {
                    if (st.length() > 0) st.append("\n");
                    st.append(p);
                }
                return st.toString();
            }

            public String toString() {
                StringBuffer ret = new StringBuffer();
                ret.append("state["+idx+"]: ");
                for(Position p : this) ret.append("{"+p+"}  ");
                return ret.toString();
            }

            public int toInt() { return idx; }
        }

        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("parse table");
            for(State<Token> state : all_states) {
                sb.append("  " + state + "\n");
                for(Topology<Token> t : state.shifts) {
                    sb.append("      shift  \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => ");
                    for(State st : state.shifts.getAll(t))
                        sb.append(st.idx+"  ");
                    sb.append("\n");
                }
                for(Topology<Token> t : state.reductions)
                    sb.append("      reduce \""+
                              new edu.berkeley.sbp.chr.CharTopology((IntegerTopology<Character>)t)+"\" => " +
                              state.reductions.getAll(t) + "\n");
                for(Sequence s : state.gotoSetNonTerminals.keySet())
                    sb.append("      goto   "+state.gotoSetNonTerminals.get(s)+" from " + s + "\n");
            }
            return sb.toString();
        }
    }

    // Helpers //////////////////////////////////////////////////////////////////////////////
    
    private static void reachable(Sequence s, HashSet<Position> h) {
        reachable(s.firstp(), h);
        //for(Sequence ss : s.needs()) reachable(ss, h);
        //for(Sequence ss : s.hates()) reachable(ss, h);
    }
    private static void reachable(Element e, HashSet<Position> h) {
        if (e instanceof Atom) return;
        for(Sequence s : ((Union)e))
            reachable(s, h);
    }
    private static void reachable(Position p, HashSet<Position> h) {
        if (h.contains(p)) return;
        h.add(p);
        if (p.element() != null) reachable(p.element(), h);
    }
    //public static Cache mastercache = null;
}
